# -*- coding: utf-8 -*-
import sys

from halo import Halo

from suite_py.lib import logger, metrics
from suite_py.lib.handler import prompt_utils
from suite_py.lib.handler.drone_handler import DroneHandler
from suite_py.lib.handler.git_handler import GitHandler
from suite_py.lib.handler.github_handler import GithubHandler


class Docker:
    # pylint: disable=too-many-instance-attributes
    def __init__(self, action, project, config, tokens, flags=None):
        self._action = action
        self._project = project
        self._flags = flags
        self._config = config
        self._tokens = tokens
        self._github = GithubHandler(tokens)
        self._repo = self._github.get_repo(project)
        self._git = GitHandler(project, config)
        self._drone = DroneHandler(config, tokens, repo=project)

    @metrics.command("docker")
    def run(self):
        if self._project != "docker":
            logger.error("`suite-py docker` must run inside docker repository.")
            sys.exit(-1)

        self._git.fetch()

        pipelines = self._drone.parse_yaml()

        if pipelines is None:
            logger.error("The file .drone.yml not found.")
            sys.exit(1)

        images = []
        for pipeline in pipelines:
            if "trigger" not in pipeline or "ref" not in pipeline["trigger"]:
                continue

            trigger_tags = pipeline["trigger"]["ref"]

            for tag in trigger_tags:
                images.append(tag.replace("refs/tags/", ""))

        image = prompt_utils.ask_choices(
            "Select an image",
            images,
        )
        versions = self._get_versions(image)

        if self._action == "release":
            print_versions(versions)
            new_version = self._ask_new_version(image)
            self._create_new_version(new_version)

        elif self._action == "versions":
            print_versions(versions)

    def _create_new_version(self, version):
        if prompt_utils.ask_confirm(f"Tag {version} will be deployed. Confirm?"):
            self._git.tag(version, version)
            self._git.push(version)

    def _ask_new_version(self, image):
        new = prompt_utils.ask_questions_input(
            "Write new version to be deployed (eg: 11.3-1):"
        )
        if new[0] == "v":
            logger.error("Insert new version without `v`")
            sys.exit(-1)
        return f'{image.replace("*", "")}{new}'

    def _get_versions(self, image):
        versions = []
        with Halo(text="Retrieving tags...", spinner="dots", color="magenta"):
            tags = self._github.get_tags("docker")
            for tag in tags:
                if image.replace("*", "") in tag.name:
                    versions.append(tag)
            return versions


def print_versions(versions):
    message = "\n".join(["* " + v.name for v in versions])

    logger.info(f"\nVersions list:\n{message}\n")



from suite_py.lib import metrics
from suite_py.lib.handler.okta_handler import Okta


class Login:
    def __init__(self, config, tokens):
        self._okta = Okta(config, tokens)

    @metrics.command("login")
    def run(self):
        self._okta.login()



import time
import typing

from suite_py.lib import logger, oauth
from suite_py.lib.config import Config
from suite_py.lib.tokens import Tokens

_SCOPE = "openid offline_access"


class Okta:
    def __init__(self, config: Config, tokens: Tokens) -> None:
        self._config = config
        self._tokens = tokens

    def login(self):
        res = oauth.authorization_code_flow(
            self._config.okta["client_id"],
            self._config.okta["base_url"],
            _SCOPE,
        )

        self._update_tokens(res)

    def get_id_token(self) -> str:
        """
        Returns an id_token, performing a token refresh if needed
        Raises on an invalid or missing refresh token
        """
        return self._get_id_token() or self._refresh()

    def _refresh(self) -> str:
        logger.debug("Refreshing id_token")

        refresh_token = self._get_refresh_token()
        if not isinstance(refresh_token, str):
            raise Exception(
                "Invalid okta refresh token. Try logging in with `suite_py login`"
            )
        res = oauth.do_refresh_token(
            self._config.okta["client_id"],
            self._config.okta["base_url"],
            _SCOPE,
            refresh_token,
        )

        return self._update_tokens(res)

    def _update_tokens(self, tokens: oauth.OAuthTokenResponse) -> str:
        if not tokens.id_token:
            raise Exception("Okta didn't return a new id_token. This shouldn't happen.")
        if not tokens.refresh_token:
            raise Exception(
                "Okta didn't return a new refresh_token. This shouldn't happen."
            )

        self._set_refresh_token(tokens.refresh_token)

        expires_at = time.time() + tokens.expires_in
        self._set_id_token(tokens.id_token, expires_at)

        return tokens.id_token

    def _get_refresh_token(self) -> typing.Optional[str]:
        return self._tokens.okta().get("refresh_token", None)

    def _set_refresh_token(self, token: str):
        okta = self._tokens.okta()
        okta["refresh_token"] = token
        self._tokens.edit("okta", okta)

    def _get_id_token(self) -> typing.Optional[str]:
        (token, expiration) = self._tokens.okta().get("id_token", (None, None))
        if token is not None and time.time() < expiration:
            return token
        return None

    def _set_id_token(self, token: str, expiration: float):
        okta = self._tokens.okta()
        okta["id_token"] = (token, expiration)
        self._tokens.edit("okta", okta)



# -*- encoding: utf-8 -*-
import base64
import marshal
import os
import time
from sys import platform

import keyring
from InquirerPy import prompt

from suite_py.lib import logger

TOKENS = ["github", "youtrack", "drone"]


def platform_has_chmod():
    return platform != "win32"


def should_use_keyring():
    # Test that the OS' keyring provider actually works.
    # If it doesn't work, or the OS doesn't provide one, we can just fall back to using a file.
    try:
        test_str = str(round(time.time() * 1000))
        keyring.set_password("suite_py_test", "test", test_str)
        test_success = keyring.get_password("suite_py_test", "test") == test_str

        try:
            keyring.delete_password("suite_py_test", "test")
        except Exception:
            # We can safely ignore this if deleting the test keyring fails
            pass

        if not test_success:
            raise Exception("Loaded keyring didn't match the saved test keyring")

        return True
    except Exception as error:
        logger.warning(
            f"suite-py was unable to store secrets using your OS' keyring provider; falling back to file-based tokens (less secure.) Error: {error}"
        )
        return False


USE_KEYRING = should_use_keyring()


class Tokens:
    def __init__(self, file_name=os.path.join(os.environ["HOME"], ".suite_py/tokens")):
        self._file_name = file_name

        try:
            self._tokens = self.load() or {}
        except Exception as error:
            self._tokens = {}
            logger.warning(f"Failed to load tokens! Error: {error}")

        self._changed = False
        self.check()
        if self._changed:
            self.save()

        self.chmod()

    def check(self):
        for token in TOKENS:
            if not self._tokens.get(token):
                self._tokens[token] = prompt(
                    [
                        {
                            "type": "input",
                            "name": token,
                            "message": f"Insert your {token.capitalize()} token:",
                        }
                    ]
                )[token]
                self._changed = True

    def load_from_keyring(self):
        decoded = keyring.get_password("suite_py", "tokens")
        if decoded is None:
            return None

        decoded = base64.b64decode(decoded)
        decoded = marshal.loads(decoded)
        return decoded

    def load_from_file(self):
        if os.path.exists(self._file_name):
            with open(self._file_name, "rb") as configfile:
                return marshal.load(configfile)
        else:
            return None

    def load(self):
        has_tokens_file = os.path.exists(self._file_name)

        if USE_KEYRING and not has_tokens_file:
            tokens = self.load_from_keyring()
        else:
            tokens = self.load_from_file()

            try:
                if USE_KEYRING and has_tokens_file and tokens is not None:
                    # If we are on a platform that uses keyring instead of a tokens file, attempt to migrate the tokens file to keyring.
                    # What we'll do is try to save the tokens to keyring, and then load them from the keyring.
                    # If the loaded tokens match the tokens in the file, we can safely delete the file, completing the migration.
                    self.save(tokens)

                    if tokens == self.load_from_keyring():
                        try:
                            os.remove(self._file_name)
                        except Exception as error:
                            logger.warning(
                                f"Could not delete old tokens file! Your tokens may be vulnerable to attackers. Please `rm {self._file_name}` manually. Error: {error}"
                            )
            except Exception as error:
                logger.warning(
                    f"Failed to migrate old tokens file to keyring! Your tokens may be vulnerable to attackers. Error: {error}"
                )

        return tokens

    def save(self, tokens=None):
        if tokens is None:
            # default `tokens` to `self._tokens`
            tokens = self._tokens

        if USE_KEYRING:
            encoded = marshal.dumps(tokens)
            encoded = base64.b64encode(encoded)
            encoded = encoded.decode("utf-8")
            keyring.set_password("suite_py", "tokens", encoded)
        else:
            with open(self._file_name, "wb") as configfile:
                marshal.dump(tokens, configfile)

    def chmod(self):
        if os.path.exists(self._file_name):
            try:
                if not platform_has_chmod():
                    raise Exception("chmod is not supported on this OS")

                os.chmod(self._file_name, 0o600)
            except Exception as error:
                logger.warning(
                    f"Could not set permissions on tokens file! Your tokens may be vulnerable to attackers. Please `chmod 600 {self._file_name}` manually. Error: {error}"
                )

    def edit(self, service, token):
        self._tokens[service] = token
        self.save()

    def keys(self):
        return self._tokens.keys()

    @property
    def github(self):
        return self._tokens["github"]

    @property
    def youtrack(self):
        return self._tokens["youtrack"]

    @property
    def drone(self):
        return self._tokens["drone"]

    def okta(self):
        return self._tokens.get("okta", {})



#! -*- encoding: utf-8 -*-

# We import command modules inside their respective function bodies
# for performance reasons, so turn off the lint warning
# pylint: disable=import-outside-toplevel

import os
import sys
from functools import wraps
from typing import Optional

import click
import pkg_resources
import requests
from autoupgrade import Package
from click.exceptions import ClickException

from suite_py.__version__ import __version__
from suite_py.commands.context import Context
from suite_py.lib import logger, metrics
from suite_py.lib.config import Config
from suite_py.lib.handler import git_handler as git
from suite_py.lib.handler import prompt_utils
from suite_py.lib.handler.captainhook_handler import CaptainHook
from suite_py.lib.handler.okta_handler import Okta
from suite_py.lib.tokens import Tokens

ALLOW_NO_GIT_SUBCOMMAND = ["login", "aggregator"]
ALLOW_NO_HOME_SUBCOMMAND = ["login", "aggregator"]


def maybe_inject_truststore() -> None:
    """
    Injects the truststore package into the system ssl store, to fix verficiation certificate issues when using warp.
    """
    if sys.version_info >= (3, 10):
        import truststore

        truststore.inject_into_ssl()
    else:
        logger.warning(
            "Your python version is older than 3.10 and doesn't support the truststore package. You might experience issues with certificate verification when connected to the warp VPN.\nPlease update to python 3.10 or newer"
        )


def fetch_latest_version() -> Optional[pkg_resources.packaging.version.Version]:
    """
    Fetches the latest version of suite-py as a pkg_resources parsed version
    Returns None on error
    """
    try:
        # pylint: disable-next=missing-timeout
        pkg_info = requests.get("https://pypi.org/pypi/suite-py/json").json()
        return pkg_resources.parse_version(pkg_info["info"]["version"])
    except Exception:
        logger.warning("Failed to fetch latest version of suite-py!")
        return None


def upgrade_suite_py_if_needed(break_on_missing_package: bool = False) -> None:
    """
    Sometimes, when `suite-py` is launched with a virtual environment active, autoupgrade
    cannot see the package as installed (as the command is used from the "global" user environment)
    and raises an exception.
    """

    try:
        pkg_resources.get_distribution("suite_py")
    except pkg_resources.DistributionNotFound as error:
        # We are in a virtual environment where suite-py is not installed, don't bother trying to upgrade
        if break_on_missing_package:
            raise error

        logger.warning(
            "Skipping Suite-Py autoupgrade because the package was not found in the current environment."
        )
        return

    current_version = pkg_resources.parse_version(__version__)

    try:
        # If available, upgrade (if needed)
        latest_version = fetch_latest_version()
        if latest_version is None or latest_version > current_version:
            # If we fail to fetch the latest version, fallback to attempting the upgrade anyway
            Package("suite_py").upgrade()
    except Exception as error:
        logger.warning(f"An error occurred whilst trying to upgrade suite-py: {error}")


# Catches any exceptions thrown and turns them into a ClickException
# So they don't print a stacktrace
def catch_exceptions(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except ClickException as e:
            raise e
        except Exception as e:
            logger.debug("An error occured:", exc_info=True)
            raise ClickException(f"{str(e)}. rerun with -v for more details") from e

    return wrapper


@click.group()
@click.option(
    "--project",
    type=click.Path(exists=True),
    default=os.getcwd(),
    help="Path of the project to run the command on (the default is current directory)",
)
@click.option(
    "--timeout",
    type=click.INT,
    help="Timeout in seconds for Captainhook operations",
)
@click.option("-v", "--verbose", count=True)
@click.pass_context
@catch_exceptions
def main(ctx, project, timeout, verbose):
    config = Config()

    logger.setup(verbose)

    logger.debug(f"v{__version__}")
    maybe_inject_truststore()
    upgrade_suite_py_if_needed(break_on_missing_package=False)

    if ctx.invoked_subcommand not in ALLOW_NO_GIT_SUBCOMMAND:
        project = git.get_root_folder(project)

    if ctx.invoked_subcommand not in ALLOW_NO_GIT_SUBCOMMAND and not git.is_repo(
        project
    ):
        raise Exception(f"the folder {project} is not a git repo")

    if ctx.invoked_subcommand not in ALLOW_NO_HOME_SUBCOMMAND and not os.path.basename(
        project
    ) in os.listdir(config.user["projects_home"]):
        raise Exception(
            f"the folder {project} is not in {config.user['projects_home']}"
        )

    skip_confirmation = False
    if type(config.user.get("skip_confirmation")).__name__ == "bool":
        skip_confirmation = config.user.get("skip_confirmation")
    elif type(
        config.user.get("skip_confirmation")
    ).__name__ == "list" and ctx.invoked_subcommand in config.user.get(
        "skip_confirmation"
    ):
        skip_confirmation = True

    if not skip_confirmation and not prompt_utils.ask_confirm(
        f"Do you want to continue on project {os.path.basename(project)}?"
    ):
        return
    if timeout:
        config.user["captainhook_timeout"] = timeout

    project = os.path.basename(project)
    tokens = Tokens()
    okta = Okta(config, tokens)
    captainhook = CaptainHook(config, okta, tokens)

    ctx.obj = Context(
        project=project,
        tokens=tokens,
        okta=okta,
        config=config,
        captainhook=captainhook,
    )

    ctx.obj.call(metrics.setup)

    # Skip chdir if not needed
    if (
        ctx.invoked_subcommand not in ALLOW_NO_GIT_SUBCOMMAND
        or ctx.invoked_subcommand not in ALLOW_NO_HOME_SUBCOMMAND
    ):
        os.chdir(os.path.join(config.user["projects_home"], ctx.obj.project))


@main.result_callback()
@click.pass_obj
def cleanup(_obj, _, **_kwargs):
    metrics.async_upload()


@main.command("bump", help="Bumps the project version based on the .versions.yml file")
@click.option("--project", required=False, type=str)
@click.option(
    "--version",
    required=False,
    type=str,
    help="Version to apply. If not passed, you will be prompted to insert or select one from a predefined list",
)
@click.pass_obj
@catch_exceptions
def bump(obj: Context, project: Optional[str] = None, version: Optional[str] = None):
    from suite_py.commands.bump import Bump

    obj.call(Bump).run(project=project, version=version)


@main.command(
    "create-branch", help="Create local branch and set the YouTrack card in progress"
)
@click.option("--card", type=click.STRING, help="YouTrack card number (ex. PRIMA-123)")
@click.pass_obj
@catch_exceptions
def cli_create_branch(obj, card):
    from suite_py.commands.create_branch import CreateBranch

    obj.call(CreateBranch, card=card).run()


@main.command("lock", help="Lock project on staging or prod")
@click.argument(
    "environment", type=click.Choice(("staging", "production", "deploy", "merge"))
)
@click.pass_obj
@catch_exceptions
def cli_lock_project(obj, environment):
    from suite_py.commands.project_lock import ProjectLock

    obj.call(ProjectLock, env=environment, action="lock").run()


@main.command("unlock", help="Unlock project on staging or prod")
@click.argument(
    "environment", type=click.Choice(("staging", "production", "deploy", "merge"))
)
@click.pass_obj
@catch_exceptions
def cli_unlock_project(obj, environment):
    from suite_py.commands.project_lock import ProjectLock

    obj.call(ProjectLock, env=environment, action="unlock").run()


@main.command("open-pr", help="Open a PR on GitHub")
@click.pass_obj
@catch_exceptions
def cli_open_pr(obj):
    from suite_py.commands.open_pr import OpenPR
    from suite_py.commands.ask_review import AskReview

    ask_review = obj.call(AskReview)
    obj.call(OpenPR, ask_review=ask_review).run()


@main.command("ask-review", help="Requests a PR review")
@click.pass_obj
@catch_exceptions
def cli_ask_review(obj):
    from suite_py.commands.ask_review import AskReview

    obj.call(AskReview).run()


@main.command(
    "merge-pr", help="Merge the selected branch to master if all checks are OK"
)
@click.pass_obj
@catch_exceptions
def cli_merge_pr(obj):
    from suite_py.commands.merge_pr import MergePR

    obj.call(MergePR).run()


@main.group("release", help="Manage releases")
def release():
    pass


@release.command(
    "create", help="Create a github release (and deploy it if GHA are used)"
)
@click.option(
    "--deploy",
    is_flag=True,
    help="Trigger deploy with Drone CI after release creation (Github Actions based microservices will be automatically deployed on production)",
)
@click.pass_obj
@catch_exceptions
def cli_release_create(obj, deploy):
    from suite_py.commands.release import Release

    obj.call(Release, action="create", flags={"deploy": deploy}).run()


@release.command("deploy", help="Deploy a github release with Drone CI")
@click.pass_obj
@catch_exceptions
def cli_release_deploy(obj):
    from suite_py.commands.release import Release

    obj.call(Release, action="deploy").run()


@release.command("rollback", help="Rollback a deployment")
@click.pass_obj
@catch_exceptions
def cli_release_rollback(obj):
    from suite_py.commands.release import Release

    obj.call(Release, "rollback").run()


@main.command("deploy", help="Deploy master branch in production")
@click.pass_obj
@catch_exceptions
def cli_deploy(obj):
    from suite_py.commands.deploy import Deploy

    obj.call(Deploy).run()


@main.group("docker", help="Manage docker images")
def docker():
    pass


@docker.command("release", help="Release new docker image")
@click.pass_obj
@catch_exceptions
def cli_docker_release(obj):
    from suite_py.commands.docker import Docker

    obj.call(Docker, action="release").run()


@docker.command("versions", help="List all available versions of specific image")
@click.pass_obj
@catch_exceptions
def cli_docker_versions(obj):
    from suite_py.commands.docker import Docker

    obj.call(Docker, action="versions").run()


@main.command("status", help="Current status of a project")
@click.pass_obj
@catch_exceptions
def cli_status(obj):
    from suite_py.commands.status import Status

    obj.call(Status).run()


@main.command("check", help="Verify authorisations for third party services")
@click.pass_obj
@catch_exceptions
def cli_check(obj):
    from suite_py.commands.check import Check

    obj.call(Check).run()


@main.command("id", help="Get the ID of the hosts where the task is running")
@click.argument("environment", type=click.Choice(("staging", "production")))
@click.pass_obj
@catch_exceptions
def cli_id(obj, environment):
    from suite_py.commands.id import ID

    obj.call(ID, environment=environment).run()


@main.command("ip", help="Get the IP addresses of the hosts where the task is running")
@click.argument("environment", type=click.Choice(("staging", "production")))
@click.pass_obj
@catch_exceptions
def cli_ip(obj, environment):
    from suite_py.commands.ip import IP

    obj.call(IP, environment=environment).run()


@main.command("generator", help="Generate different files from templates")
@click.pass_obj
@catch_exceptions
def cli_generator(obj):
    from suite_py.commands.generator import Generator

    obj.call(Generator).run()


@main.group(
    "aggregator",
    help="Manage CNAMEs of aggregators in QA envs",
    invoke_without_command=True,
)
@click.option(
    "-l", "--list", "show_list", help="deprecated", required=False, count=True
)
@click.option("-c", "--change", "change", help="deprecated", required=False, count=True)
@click.pass_context
def aggregator(ctx, show_list, change):
    if show_list == 1 or change == 1:
        logger.error(
            "suite-py aggregator [-c|-l] has been hard-deprecated. Use suite-py aggregator [list|change] instead."
        )
        sys.exit(0)

    if ctx.invoked_subcommand is None:
        logger.error("Missing command. Try suite-py aggregator --help for help.")
        sys.exit(0)


@aggregator.command("list", help="List all aggregators with the current record")
@click.pass_obj
@catch_exceptions
def cli_aggregator_list(obj):
    from suite_py.commands.aggregator import Aggregator

    obj.call(Aggregator, action="list").run()


@aggregator.command("change", help="Change aggregator record")
@click.pass_obj
@catch_exceptions
def cli_aggregator_change(obj):
    from suite_py.commands.aggregator import Aggregator

    obj.call(Aggregator, action="list").run()


@main.command("login", help="manage login against Auth0")
@click.pass_obj
@catch_exceptions
def login(obj):
    from suite_py.commands.login import Login

    obj.call(Login).run()


@main.group(
    "secret", help="Manage secrets grants in multiple countries (aws-vault needed)"
)
def secret():
    pass


@secret.command("create", help="Create a new secret")
@click.option("-b", "--base-profile", "base_profile", required=False)
@click.option("-f", "--secret-file", "secret_file", required=False)
@click.pass_obj
@catch_exceptions
def cli_secret_create(obj, base_profile, secret_file):
    from suite_py.commands.secret import Secret

    obj.call(
        Secret, action="create", base_profile=base_profile, secret_file=secret_file
    ).run()


@secret.command("grant", help="Grant permissions to an existing secret")
@click.option("-b", "--base-profile", "base_profile", required=False)
@click.option("-f", "--secret-file", "secret_file", required=False)
@click.pass_obj
@catch_exceptions
def cli_secret_grant(obj, base_profile, secret_file):
    from suite_py.commands.secret import Secret

    obj.call(
        Secret, action="grant", base_profile=base_profile, secret_file=secret_file
    ).run()


@main.command("batch-job", help="Run batch job on kube")
@click.argument("environment", type=click.Choice(("staging", "production")))
@click.option(
    "-c",
    "--cpu",
    "cpu_request",
    required=True,
    type=str,
    help="Millicpu format. E.g: 1000m (=1CPU)",
)
@click.option(
    "-m",
    "--memory",
    "memory_request",
    required=True,
    type=str,
    help="Mebibytes or gibibyte format. E.g: 256Mi or 1Gi",
)
@click.pass_obj
@catch_exceptions
def cli_run_batch_job(obj, environment, cpu_request, memory_request):
    from suite_py.commands.batch_job import BatchJob

    obj.call(
        BatchJob,
        environment=environment,
        cpu_request=cpu_request,
        memory_request=memory_request,
    ).run()


@main.command("set-token", help="Create or update a service token")
@click.pass_obj
@catch_exceptions
def cli_set_token(obj):
    from suite_py.commands.set_token import SetToken

    obj.call(SetToken).run()



# -*- encoding: utf-8 -*-
import functools
from suite_py.lib.config import Config

from suite_py.lib.handler.captainhook_handler import CaptainHook
from suite_py.lib.handler.metrics_handler import Metrics

_metrics_handler = None


def _metrics() -> Metrics:
    if _metrics_handler:
        return _metrics_handler

    raise Exception(
        "command_executed called before logger.setup(). This is a bug, please report it"
    )


def setup(config: Config, captainhook: CaptainHook):
    global _metrics_handler
    _metrics_handler = Metrics(config=config, captainhook=captainhook)


def command_executed(command):
    _metrics().command_executed(command)


def async_upload():
    _metrics().async_upload()


# Decorator that emits the command_executed metric with the given command name,
# and sets the success paramter to false if the function exited by throwing an error
def command(command: str):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                res = func(*args, **kwargs)
                _metrics().command_executed(command, success=True)
                return res
            except Exception as e:
                _metrics().command_executed(command, success=False)
                raise e
            # We use sys.exit everywhere, report those
            except SystemExit as e:
                # Report sys.exit(0) and sys.exit() as successes
                success = e.code in (0, None)
                _metrics().command_executed(command, success=success)

                raise e

        return wrapper

    return decorator
